#!/usr/bin/env python
# _*_ coding: utf-8 _*_

import sys, os
pPath = os.path.split(os.path.realpath(__file__))[0]
sys.path.append(pPath)
from PyQt5.QtWidgets import (QApplication, QMainWindow, QWidget, QDesktopWidget, QLabel, QHBoxLayout, QMessageBox, QAction, QFileDialog)
from PyQt5.QtGui import QIcon, QFont, QPixmap, QCloseEvent, QDesktopServices
from PyQt5.QtCore import Qt, QUrl
from util import Modules, InputDialog, PlotWidgets, MachineLearning
import iLearnPlusBasic, iLearnPlusEstimator, iLearnPlusAutoML, iLearnPlusLoadModel
import qdarkstyle
import threading
import pandas as pd
import numpy as np

class MainWindow(QMainWindow):
    def __init__(self):
        super(MainWindow, self).__init__()
        self.initUI()

    def initUI(self):
        # initialize GUI
        self.setWindowTitle('iLearnPlus')
        # self.resize(750, 500)
        self.setMaximumSize(600, 400)
        self.setMinimumSize(600, 400)

        self.setWindowIcon(QIcon(os.path.join(pPath, 'images', 'logo.ico')))
        self.setFont(QFont('Arial'))
        bar = self.menuBar()
        app = bar.addMenu('Applications')
        basic = QAction('iLearnPlus Basic', self)
        basic.triggered.connect(self.openBasicWindow)
        estimator = QAction('iLearnPlus Estimator', self)
        estimator.triggered.connect(self.openEstimatorWindow)
        autoML = QAction('iLearnPlus AutoML', self)
        autoML.triggered.connect(self.openMLWindow)
        loadModel = QAction('Load model(s)', self)
        loadModel.triggered.connect(self.openLoadModelWindow)
        quit = QAction('Exit', self)
        quit.triggered.connect(self.closeEvent)
        app.addAction(basic)
        app.addAction(estimator)
        app.addAction(autoML)
        app.addSeparator()
        app.addAction(loadModel)
        app.addSeparator()
        app.addAction(quit)

        visual = bar.addMenu('Visualization')
        roc = QAction('Plot ROC curve', self)
        roc.triggered.connect(lambda: self.plotCurve('ROC'))
        prc = QAction('Plot PRC curve', self)
        prc.triggered.connect(lambda: self.plotCurve('PRC'))
        boxplot = QAction('Boxplot', self)
        boxplot.triggered.connect(self.drawBoxplot)
        heatmap = QAction('Heatmap', self)
        heatmap.triggered.connect(self.drawHeatmap)
        scatterPlot = QAction('Scatter plot', self)
        scatterPlot.triggered.connect(self.scatterPlot)
        data = QAction('Distribution visualization', self)
        data.triggered.connect(self.displayHist)
        visual.addActions([roc, prc, boxplot, heatmap, scatterPlot, data])

        tools = bar.addMenu('Tools')        
        fileTF = QAction('File format transformation', self)
        fileTF.triggered.connect(self.openFileTF)
        mergeFile = QAction('Merge feature set files into one', self)
        mergeFile.triggered.connect(self.mergeCodingFiles)
        tools.addActions([fileTF, mergeFile])

        help = bar.addMenu('Help')
        document = QAction('Document', self)
        document.triggered.connect(self.openDocumentUrl)
        about = QAction('About', self)
        about.triggered.connect(self.openAbout)
        help.addActions([document, about])

        # move window to center
        self.moveCenter()

        self.widget = QWidget()
        hLayout = QHBoxLayout(self.widget)
        hLayout.setAlignment(Qt.AlignCenter)
        label = QLabel()
        # label.setMaximumWidth(600)
        label.setPixmap(QPixmap(os.path.join(pPath, 'images', 'logo.png')))
        hLayout.addWidget(label)
        self.setCentralWidget(self.widget)

    def moveCenter(self):
        screen = QDesktopWidget().screenGeometry()
        size = self.geometry()
        newLeft = (screen.width() - size.width()) / 2
        newTop = (screen.height() - size.height()) / 2
        self.move(int(newLeft), int(newTop))

    def openBasicWindow(self):
        self.basicWin = iLearnPlusBasic.ILearnPlusBasic()
        self.basicWin.setFont(QFont('Arial', 10))
        self.basicWin.setStyleSheet(qdarkstyle.load_stylesheet_pyqt5())
        self.basicWin.close_signal.connect(self.recover)
        self.basicWin.show()
        self.setDisabled(True)
        self.setVisible(False)

    def openEstimatorWindow(self):
        self.estimatorWin = iLearnPlusEstimator.ILearnPlusEstimator()
        self.estimatorWin.setFont(QFont('Arial', 10))
        self.estimatorWin.setStyleSheet(qdarkstyle.load_stylesheet_pyqt5())
        self.estimatorWin.close_signal.connect(self.recover)
        self.estimatorWin.show()
        self.setDisabled(True)
        self.setVisible(False)

    def openMLWindow(self):
        self.mlWin = iLearnPlusAutoML.ILearnPlusAutoML()
        self.mlWin.setFont(QFont('Arial', 10))
        self.mlWin.setStyleSheet(qdarkstyle.load_stylesheet_pyqt5())
        self.mlWin.close_signal.connect(self.recover)
        self.mlWin.show()
        self.setDisabled(True)
        self.setVisible(False)

    def openLoadModelWindow(self):
        self.loadWin = iLearnPlusLoadModel.iLearnPlusLoadModel()
        # self.loadWin.setFont(QFont('Arial', 10))
        self.loadWin.setStyleSheet(qdarkstyle.load_stylesheet_pyqt5())
        self.loadWin.close_signal.connect(self.recover)
        self.loadWin.show()
        self.setDisabled(True)
        self.setVisible(False)

    def closeEvent(self, event):
        reply = QMessageBox.question(self, 'Confirm Exit', 'Are you sure want to quit iLearnPlus?', QMessageBox.Yes | QMessageBox.No, QMessageBox.No)
        if reply == QMessageBox.Yes:
            sys.exit(0)
        else:
            if event:
                event.ignore()

    def plotCurve(self, curve='ROC'):
        self.curveWin = Modules.PlotCurve(curve)
        self.curveWin.setStyleSheet(qdarkstyle.load_stylesheet_pyqt5())
        self.curveWin.show()

    def displayHist(self):
        self.hist = Modules.Hist()
        self.hist.setStyleSheet(qdarkstyle.load_stylesheet_pyqt5())
        self.hist.show()

    def scatterPlot(self):
        self.scatter = Modules.ScatterPlot()
        self.scatter.setStyleSheet(qdarkstyle.load_stylesheet_pyqt5())
        self.scatter.show()

    def openFileTF(self):
        try:
            fileName, ok = InputDialog.QFileTransformation.getValues()
            if ok:
                kw = {}
                TFData = MachineLearning.ILearnMachineLearning(kw)            
                TFData.load_data(fileName, target='Training')
                if not TFData.training_dataframe is None:
                    saved_file, ok = QFileDialog.getSaveFileName(self, 'Save to', os.path.join(pPath, 'data'), 'CSV Files (*.csv);;TSV Files (*.tsv);;SVM Files(*.svm);;Weka Files (*.arff)')
                    if ok:
                        ok1 = TFData.save_coder(saved_file, 'training')
                        if not ok1:
                            QMessageBox.critical(self, 'Error', str(self.TFData.error_msg), QMessageBox.Ok | QMessageBox.No, QMessageBox.Ok)
        except Exception as e:
            QMessageBox.critical(self, 'Error', str(e), QMessageBox.Ok | QMessageBox.No, QMessageBox.Ok)

    def mergeCodingFiles(self):
        try:
            coding_files, ok = QFileDialog.getOpenFileNames(self, 'Open coding files (more file can be selected)', os.path.join(pPath, 'data'), 'CSV Files (*.csv);;TSV Files (*.tsv);;SVM Files(*.svm);;Weka Files (*.arff)')
            merged_codings = None
            labels = None
            if len(coding_files) > 0:
                dataframe, datalabel = None, None
                for file in coding_files:
                    if file.endswith('.tsv'):
                        df = pd.read_csv(file, sep='\t', header=None)
                        dataframe = df.iloc[:, 1:]
                        dataframe.index=['Sample_%s'%i for i in range(dataframe.values.shape[0])]
                        dataframe.columns = ['F_%s'%i for i in range(dataframe.values.shape[1])]
                        datalabel = np.array(df.iloc[:, 0]).astype(int)
                    elif file.endswith('.csv'):
                        df = pd.read_csv(file, sep=',', header=None)
                        dataframe = df.iloc[:, 1:]
                        dataframe.index=['Sample_%s'%i for i in range(dataframe.values.shape[0])]
                        dataframe.columns = ['F_%s'%i for i in range(dataframe.values.shape[1])]
                        datalabel = np.array(df.iloc[:, 0]).astype(int)
                    elif file.endswith('.svm'):
                        with open(file) as f:
                            record = f.read().strip()
                        record = re.sub('\d+:', '', record)
                        array = np.array([[i for i in item.split()] for item in record.split('\n')])
                        dataframe = pd.DataFrame(array[:, 1:], dtype=float)
                        dataframe.index=['Sample_%s'%i for i in range(dataframe.values.shape[0])]
                        dataframe.columns = ['F_%s'%i for i in range(dataframe.values.shape[1])]
                        datalabel = array[:, 0].astype(int)
                    else:
                        with open(file) as f:
                            record = f.read().strip().split('@')[-1].split('\n')[1:]
                        array = np.array([item.split(',') for item in record])
                        dataframe = pd.DataFrame(array[:, 0:-1], dtype=float)
                        dataframe.index=['Sample_%s'%i for i in range(dataframe.values.shape[0])]
                        dataframe.columns = ['F_%s'%i for i in range(dataframe.values.shape[1])]
                        label = []
                        for i in array[:, -1]:
                            if i == 'yes':
                                label.append(1)
                            else:
                                label.append(0)
                        datalabel = np.array(label)
                
                    if merged_codings is None:
                        merged_codings = np.hstack((datalabel.reshape((-1, 1)), dataframe.values))
                    else:
                        merged_codings = np.hstack((merged_codings, dataframe.values))
            if merged_codings is not None:
                saved_file, ok = QFileDialog.getSaveFileName(self, 'Save to', './data', 'CSV Files (*.csv);;TSV Files (*.tsv);;SVM Files(*.svm);;Weka Files (*.arff)')
                data = merged_codings
                if saved_file.endswith('.csv'):
                    np.savetxt(saved_file, data, fmt="%s", delimiter=',')
                if saved_file.endswith('.tsv'):
                    np.savetxt(saved_file, data, fmt="%s", delimiter='\t')
                if saved_file.endswith('.svm'):
                    with open(saved_file, 'w') as f:
                        for line in data:
                            f.write('%s' % line[0])
                            for i in range(1, len(line)):
                                f.write('  %d:%s' % (i, line[i]))
                            f.write('\n')
                if saved_file.endswith('.arff'):
                    with open(saved_file, 'w') as f:
                        f.write('@relation descriptor\n\n')
                        for i in range(1, len(data[0])):
                            f.write('@attribute f.%d numeric\n' % i)
                        f.write('@attribute play {yes, no}\n\n')
                        f.write('@data\n')
                        for line in data:
                            for fea in line[1:]:
                                f.write('%s,' % fea)
                            if int(line[0]) == 1:
                                f.write('yes\n')
                            else:
                                f.write('no\n')
        except Exception as e:
            QMessageBox.critical(self, 'Error', str(e), QMessageBox.Ok | QMessageBox.No, QMessageBox.Ok)

    def openDocumentUrl(self):
        QDesktopServices.openUrl(QUrl('https://ilearnplus.erc.monash.edu/'))

    def openAbout(self):
        QMessageBox.information(self, 'iLearnPlus', 'Version: 1.1\nAuthor: Zhen Chen\nE-mail: chenzhen-win2009@163.com', QMessageBox.Ok | QMessageBox.No, QMessageBox.Ok)

    def drawBoxplot(self):
        try:
            x, y, data, ok = InputDialog.QBoxPlotInput.getValues()
            if ok:
                self.boxWin = PlotWidgets.CustomSingleBoxplotWidget(data, x, y)
                self.boxWin.show()
        except Exception as e:
            QMessageBox.critical(self, 'Error', str(e), QMessageBox.Ok | QMessageBox.No, QMessageBox.Ok)

    def drawHeatmap(self):
        try:
            x, y, data, ok = InputDialog.QHeatmapInput.getValues()
            if ok:
                self.heatWin = PlotWidgets.CustomHeatmapWidget(data, x, y)
                self.heatWin.show()
        except Exception as e:
            QMessageBox.critical(self, 'Error', str(e), QMessageBox.Ok | QMessageBox.No, QMessageBox.Ok)

    def recover(self, module):
        try:
            if module == 'Basic':                
                del self.basicWin               
            elif module == 'Estimator':
                del self.estimatorWin
            elif module == 'AutoML':
                del self.mlWin
            elif module == 'LoadModel':
                del self.loadWin
            else:
                pass
        except Exception as e:
            pass
        self.setDisabled(False)
        self.setVisible(True)


def runiLearnPlus():
    app = QApplication(sys.argv)
    window = MainWindow()
    window.setStyleSheet(qdarkstyle.load_stylesheet_pyqt5())
    window.show()
    sys.exit(app.exec_())

if __name__ == '__main__':
    app = QApplication(sys.argv)
    window = MainWindow()
    window.setStyleSheet(qdarkstyle.load_stylesheet_pyqt5())
    window.show()
    sys.exit(app.exec_())


#!/usr/bin/env python
# _*_ coding: utf-8 _*_

import sys, os
pPath = os.path.split(os.path.realpath(__file__))[0]
sys.path.append(pPath)
from PyQt5.QtWidgets import (QApplication, QTableWidget, QWidget, QVBoxLayout, QHBoxLayout, QPushButton, QLabel,
                             QLineEdit, QTableWidgetItem, QMessageBox, QComboBox, QSpacerItem, QSizePolicy, QAbstractItemView)
from PyQt5.QtGui import QFont
from PyQt5.QtCore import pyqtSignal
import pandas as pd
import numpy as np
from PlotWidgets import HistogramWidget
import qdarkstyle

class TableWidget(QWidget):
    control_signal = pyqtSignal(list)

    def __init__(self):
        super(TableWidget, self).__init__()
        self.data = None
        self.row = 50
        self.column = 0
        self.page_dict = {}
        self.final_page = ''
        self.header = []
        self.selected_column = None
        self.initUI()

    def initUI(self):
        self.control_signal.connect(self.page_controller)
        layout = QVBoxLayout(self)
        self.tableWidget = QTableWidget()
        self.tableWidget.setEditTriggers(QAbstractItemView.NoEditTriggers)
        self.tableWidget.horizontalHeader().sectionClicked.connect(self.horizontal_header_clicked)
        self.tableWidget.setFont(QFont('Arial', 8))
        layout.addWidget(self.tableWidget)

        control_layout = QHBoxLayout()
        homePage = QPushButton("Home")
        prePage = QPushButton("< Previous")
        self.curPage = QLabel("1")
        nextPage = QPushButton("Next >")
        finalPage = QPushButton("Final")
        self.totalPage = QLabel("Total")
        skipLable_0 = QLabel("Jump to")
        self.skipPage = QLineEdit()
        self.skipPage.setMaximumWidth(50)
        confirmSkip = QPushButton("OK")
        spacer = QSpacerItem(20, 10, QSizePolicy.Expanding, QSizePolicy.Minimum)
        self.label_comboBox = QComboBox()
        # self.label_comboBox.setMaximumWidth(50)
        self.label_comboBox.addItem('All')
        self.label_comboBox.setCurrentIndex(0)
        self.label_comboBox.setToolTip('Select the sample category used to draw the histogram.')
        self.display_hist = QPushButton(' Histogram ')
        self.display_hist.clicked.connect(self.display_histogram)

        control_layout.addStretch(1)
        homePage.clicked.connect(self.__home_page)
        prePage.clicked.connect(self.__pre_page)
        nextPage.clicked.connect(self.__next_page)
        finalPage.clicked.connect(self.__final_page)
        confirmSkip.clicked.connect(self.__confirm_skip)
        control_layout.addWidget(homePage)
        control_layout.addWidget(prePage)
        control_layout.addWidget(self.curPage)
        control_layout.addWidget(nextPage)
        control_layout.addWidget(finalPage)
        control_layout.addWidget(self.totalPage)
        control_layout.addWidget(skipLable_0)
        control_layout.addWidget(self.skipPage)
        control_layout.addWidget(confirmSkip)
        control_layout.addItem(spacer)
        # control_layout.addWidget(self.label_comboBox)
        control_layout.addWidget(self.display_hist)
        control_layout.addStretch(1)
        layout.addLayout(control_layout)

    def setRowAndColumns(self, row, column):
        self.row = row
        self.column = column
        self.tableWidget.setRowCount(row)
        self.tableWidget.setColumnCount(column)

    def init_data(self, header, data):
        self.tableWidget.setHorizontalHeaderLabels(header)
        self.header = header
        self.data = data
        page = 1
        for i in range(0, self.data.shape[0], self.row):
            end = i + self.row if i + self.row < self.data.shape[0] else self.data.shape[0]
            self.page_dict[str(page)] = (i, end)
            self.final_page = str(page)
            page += 1
        self.curPage.setText('1')
        self.display_table('1')
        self.totalPage.setText('Total page: %s ' %self.final_page)

    def __home_page(self):
        self.control_signal.emit(["home", self.curPage.text()])

    def __pre_page(self):
        self.control_signal.emit(["pre", self.curPage.text()])

    def __next_page(self):
        self.control_signal.emit(["next", self.curPage.text()])

    def __final_page(self):
        self.control_signal.emit(["final", self.curPage.text()])

    def __confirm_skip(self):
        self.control_signal.emit(["confirm", self.skipPage.text()])

    def page_controller(self, signal):
        try:
            if not self.data is None:
                if 'home' == signal[0] and signal[1] != '':
                    self.display_table('1')
                    self.curPage.setText('1')
                elif 'final' == signal[0] and signal[1] != '':
                    self.display_table(self.final_page)
                    self.curPage.setText(self.final_page)
                elif 'pre' == signal[0] and signal[1] != '':
                    page = int(signal[1]) - 1 if int(signal[1]) - 1 > 0 else 1
                    self.curPage.setText(str(page))
                    self.display_table(str(page))
                elif 'next' == signal[0] and signal[1] != '':
                    page = int(signal[1]) + 1 if int(signal[1]) + 1 <= int(self.final_page) else int(self.final_page)
                    self.curPage.setText(str(page))
                    self.display_table(str(page))
                elif "confirm" == signal[0] and signal[1] != '':
                    if 1 <= int(signal[1]) <= int(self.final_page):
                        self.curPage.setText(signal[1])
                        self.display_table(signal[1])
        except Exception as e:
            pass

    def display_table(self, page):
        if page in self.page_dict:
            tmp_data = self.data[self.page_dict[page][0]: self.page_dict[page][1]]
            self.tableWidget.setRowCount(tmp_data.shape[0])
            self.tableWidget.setColumnCount(tmp_data.shape[1])
            self.tableWidget.setHorizontalHeaderLabels(self.header)
            for i in range(tmp_data.shape[0]):
                for j in range(tmp_data.shape[1]):
                    if j == 0:
                        cell = QTableWidgetItem(str(tmp_data[i][j]))
                    else:
                        cell = QTableWidgetItem(str(round(float(tmp_data[i][j]), 6)))
                    self.tableWidget.setItem(i, j, cell)
        else:
            QMessageBox.critical(self, 'Error', 'Page number out of index.', QMessageBox.Ok | QMessageBox.No, QMessageBox.Ok)
        labels = ['All'] + [str(i) for i in set(self.data[1:, 1])]
        self.label_comboBox.clear()
        self.label_comboBox.addItems(labels)
        self.label_comboBox.setCurrentIndex(0)

    def horizontal_header_clicked(self, index):
        if index == 0:
            self.selected_column = None
        else:
            self.selected_column = index

    def display_histogram(self):
        if self.selected_column is None:
            QMessageBox.critical(self, 'Error', 'Please select a column (except col 0)).', QMessageBox.Ok | QMessageBox.No, QMessageBox.Ok)
        elif self.selected_column > len(self.header) - 1:
            QMessageBox.critical(self, 'Error', 'Incorrect column index.', QMessageBox.Ok | QMessageBox.No, QMessageBox.Ok)
        else:
            # labels = np.array(self.data[1:, 1]).astype(int)
            # selected_label = self.label_comboBox.currentText()
            # data = np.array(self.data[1:, self.selected_column])
            # if selected_label == 'All':
            #     plot_data = data
            # else:
            #     plot_data = data[np.where(labels == int(selected_label))]
            data = np.hstack((self.data[1:, 1].reshape((-1, 1)), self.data[1:, self.selected_column].reshape((-1, 1)))).astype(float)
            self.hist = HistogramWidget()
            self.hist.init_data(self.header[self.selected_column], data)
            self.hist.setStyleSheet(qdarkstyle.load_stylesheet_pyqt5())
            self.hist.show()

class TableWidgetForSelPanel(TableWidget):
    def __init__(self):
        super(TableWidgetForSelPanel, self).__init__()

    def display_table(self, page):
        if page in self.page_dict:
            tmp_data = self.data[self.page_dict[page][0]: self.page_dict[page][1]]
            self.tableWidget.setRowCount(tmp_data.shape[0])
            self.tableWidget.setColumnCount(tmp_data.shape[1])
            self.tableWidget.setHorizontalHeaderLabels(self.header)
            for i in range(tmp_data.shape[0]):
                for j in range(tmp_data.shape[1]):
                    if j == 0:
                        cell = QTableWidgetItem(str(tmp_data[i][j]))
                    else:
                        cell = QTableWidgetItem(str(round(float(tmp_data[i][j]), 6)))
                    self.tableWidget.setItem(i, j, cell)
        else:
            QMessageBox.critical(self, 'Error', 'Page number out of index.', QMessageBox.Ok | QMessageBox.No, QMessageBox.Ok)
        labels = ['All'] + [str(i) for i in set(self.data[1:, 0])]
        self.label_comboBox.clear()
        self.label_comboBox.addItems(labels)
        self.label_comboBox.setCurrentIndex(0)

    def horizontal_header_clicked(self, index):
        self.selected_column = index

    def display_histogram(self):
        if self.selected_column is None:
            QMessageBox.critical(self, 'Error', 'Please select a column.', QMessageBox.Ok | QMessageBox.No, QMessageBox.Ok)
        elif self.selected_column > len(self.header) - 1:
            QMessageBox.critical(self, 'Error', 'Incorrect column index.', QMessageBox.Ok | QMessageBox.No, QMessageBox.Ok)
        else:
            # labels = np.array(self.data[:, 0]).astype(int)
            # selected_label = self.label_comboBox.currentText()
            # data = np.array(self.data[:, self.selected_column])
            # if selected_label == 'All':
            #     plot_data = data
            # else:
            #     plot_data = data[np.where(labels == int(selected_label))]
            data = np.hstack((self.data[:, 0].reshape((-1, 1)), self.data[:, self.selected_column].reshape((-1, 1)))).astype(float)
            self.hist = HistogramWidget()
            self.hist.init_data(self.header[self.selected_column], data)
            self.hist.setStyleSheet(qdarkstyle.load_stylesheet_pyqt5())
            self.hist.show()



if __name__ == '__main__':
    df = pd.read_csv('mutitask.tsv', delimiter='\t', header=0)
    app = QApplication(sys.argv)
    win = TableWidget()
    win.setRowAndColumns(40, df.values.shape[1])
    win.setTabletHeader(df.columns)
    win.init_data(df.values)

    win.show()
    sys.exit(app.exec_())



#!/usr/bin/env python
# _*_ coding: utf-8 _*_

import os, sys
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.utils.data import Dataset
import numpy as np
import math
import time


class DealDataset(Dataset):
    def __init__(self, np_data, np_label):
        self.X = torch.from_numpy(np_data)
        self.y = torch.from_numpy(np_label).view(-1, 1)
        self.len = np_data.shape[0]

    def __getitem__(self, index):
        return self.X[index], self.y[index]

    def __len__(self):
        return self.len


class Attention(nn.Module):
    """Attention layer
    Args:
        in_features_2d: the input shape for a sample. e.g. (sequence_length, dimension)
    """
    def __init__(self, in_features_2d, hidden = 10):
        super(Attention, self).__init__()
        self.input_dim = in_features_2d[1]
        self.input_length = in_features_2d[0]
        self.hidden = hidden
        self.W0 = nn.Parameter(torch.Tensor(self.input_dim, self.hidden))
        self.b0 = nn.Parameter(torch.Tensor(self.hidden, ))
        self.W  = nn.Parameter(torch.Tensor(self.hidden, 1))
        self.b  = nn.Parameter(torch.Tensor(1,))
        self.reset_parameters()

    def reset_parameters(self):
        stdv_W0 = 1./ math.sqrt(self.W0.size(0))
        self.W0.data.uniform_(-stdv_W0, stdv_W0)
        stdv_b0 = 1./ math.sqrt(self.b0.size(0))
        self.b0.data.uniform_(-stdv_b0, stdv_b0)
        stdv_W = 1. / math.sqrt(self.W.size(0))
        self.W.data.uniform_(-stdv_W, stdv_W)
        stdv_b = 1. / math.sqrt(self.b.size(0))
        self.b.data.uniform_(-stdv_b, stdv_b)

    def forward(self, x):
        """
        :param x: size: (batch_size, sequence_length, input_dim)
        :return: size: (batch_size, sequence_length + input_dim)
        """
        energy = torch.bmm(x, self.W0.repeat(x.size(0), 1, 1)) + self.b0.repeat(x.size(0), 1, 1)
        energy = torch.bmm(energy, self.W.repeat(x.size(0), 1, 1)) + self.b.repeat(x.size(0), 1, 1)
        energy = energy.view(-1, self.input_length)
        energy = F.softmax(energy, 1)

        xx = torch.bmm(torch.unsqueeze(energy, 1), x)
        output = torch.cat((torch.squeeze(xx, 1), energy), 1)

        return output


class ResidualBlock(nn.Module):
    def __init__(self, in_channel, out_channel, stride=1, shortcut=None):
        super(ResidualBlock, self).__init__()
        self.left=nn.Sequential(
            nn.Conv1d(in_channel, out_channel, kernel_size=3, stride=stride, padding=1, bias=False),
            nn.BatchNorm1d(out_channel),
            nn.ReLU(inplace=True),
            nn.Conv1d(out_channel, out_channel, kernel_size=3, stride=1, padding=1, bias=False),
            nn.BatchNorm1d(out_channel)
        )
        self.right=shortcut

    def forward(self, X):
        out = self.left(X)
        residual = X if self.right is None else self.right(X)
        out += residual
        return F.relu(out)


class Net_CNN_1(nn.Module):
    """
        Simple CNN network.
        Four Conv1d layers
        A fc layer
        A output layer
    """
    def __init__(self, device, category, input_size, sequence_length, out_channel=64, padding=2, conv_kernel_size=5,
                 pool_kernel_size=2, dense_size=64, dropout=0.5):
        super(Net_CNN_1, self).__init__()
        self.device = device
        self.input_size = input_size  # the vector dimension for each AA or Nucleotide
        self.sequence_length = sequence_length  # biological sequence length
        self.conv1 = nn.Sequential(  # (input_size sequence_length)
            nn.Conv1d(in_channels=input_size, out_channels=out_channel, kernel_size=conv_kernel_size, stride=1,
                      padding=padding),
            nn.ReLU(),
            nn.MaxPool1d(kernel_size=pool_kernel_size)  # -> (out_channel, sequence_length//2)
        )
        self.dropout1 = nn.Dropout(dropout)
        self.conv2 = nn.Sequential(
            nn.Conv1d(in_channels=out_channel, out_channels=out_channel, kernel_size=conv_kernel_size, stride=1,
                      padding=padding),
            nn.ReLU(),
            nn.MaxPool1d(kernel_size=pool_kernel_size)  # -> (out_channel, sequence_length//4)
        )
        self.dropout2 = nn.Dropout(dropout)
        self.conv3 = nn.Sequential(
            nn.Conv1d(in_channels=out_channel, out_channels=out_channel, kernel_size=conv_kernel_size, stride=1,
                      padding=padding),
            nn.ReLU(),
            nn.MaxPool1d(kernel_size=pool_kernel_size)  # -> (out_channel, sequence_length//8)
        )
        self.dropout3 = nn.Dropout(dropout)
        self.conv4 = nn.Sequential(
            nn.Conv1d(in_channels=out_channel, out_channels=out_channel, kernel_size=conv_kernel_size, stride=1,
                      padding=padding),
            nn.ReLU(),
            nn.MaxPool1d(kernel_size=pool_kernel_size)  # -> (out_channel, sequence_length//16)
        )
        self.dropout4 = nn.Dropout(dropout)
        self.fc1 = nn.Linear(out_channel * self.calculate_dimension(sequence_length, padding, conv_kernel_size),
                             dense_size)
        self.fc2 = nn.Linear(dense_size, category)

        # best model save
        self.best_model_dict = None

    def calculate_dimension(self, sequence_length, padding, kernel_size):
        L_conv1 = (sequence_length + 2 * padding - kernel_size + 1) // 2
        L_conv2 = (L_conv1 + 2 * padding - kernel_size + 1) // 2
        L_conv3 = (L_conv2 + 2 * padding - kernel_size + 1) // 2
        L_conv4 = (L_conv3 + 2 * padding - kernel_size + 1) // 2
        return L_conv4

    def forward(self, X, is_train=False):
        out = X.view(-1, self.sequence_length, self.input_size)
        out = out.permute(0, 2, 1)
        out = self.conv1(out)
        if is_train:
            out = self.dropout1(out)
        out = self.conv2(out)
        if is_train:
            out = self.dropout2(out)
        out = self.conv3(out)
        if is_train:
            out = self.dropout3(out)
        out = self.conv4(out)
        if is_train:
            out = self.dropout4(out)
        out = out.view(out.size(0), -1)
        out = F.relu(self.fc1(out))        
        # out = torch.softmax(self.fc2(out), dim=1)
        out = self.fc2(out)
        return out

    def fit(self, train_loader, valid_loader, epochs=1000, early_stopping=100, lr=0.001):
        # Loss and Optimizer
        criterion = nn.CrossEntropyLoss()
        optimizer = torch.optim.Adam(self.parameters(), lr=lr)
        minmum_loss = np.inf
        last_saving = 0
        for epoch in range(epochs):
            time0 = time.time()
            loss_train = np.inf
            for i, (batch_X, batch_y) in enumerate(train_loader):
                Xb = batch_X.float().to(self.device)
                yb = batch_y.view(-1).to(self.device)

                # Forward + Backward + Optimize
                optimizer.zero_grad()
                y_ = self.forward(Xb, is_train=True)
                loss = criterion(y_, yb.long())
                loss.backward()
                optimizer.step()
                loss_train = loss.item()
            loss_valid = float(self.evaluate(valid_loader))
            if loss_valid < minmum_loss:
                minmum_loss = loss_valid
                # torch.save(self.state_dict(), out)
                self.best_model_dict = self.state_dict()
                last_saving = epoch
            # if last_saving == epoch:
            #     print('Epoch [%d/%d]: val_loss improved to %f, saving model to model file'
            #           % (epoch + 1, epochs, minmum_loss))
            # else:
            #     print('Epoch [%d/%d]: val_loss did not improve' % (epoch + 1, epochs))
            # interval = time.time() - time0
            # print('%.1fs - loss: %f - val_loss: %f\n' % (interval, loss_train, loss_valid))
            if epoch - last_saving > early_stopping: break

        if not self.best_model_dict is None:
            self.load_state_dict(self.best_model_dict)
            # torch.save(self, 'model.pkl')

    def evaluate(self, loader):
        loss = 0
        criterion = nn.CrossEntropyLoss()
        for batch_X, batch_y in loader:
            Xb = batch_X.float().to(self.device)
            yb = batch_y.view(-1).to(self.device)
            y_ = self.forward(Xb)
            loss += criterion(y_, yb.long()).item()
        return loss / len(loader)

    def predict(self, loader):
        score = []
        for batch_X, batch_y in loader:
            Xb = batch_X.float().to(self.device)
            yb = batch_y.view(-1).to(self.device)
            y_ = self.forward(Xb)
            score.append(y_.cpu().data)
        return torch.cat(score).numpy()


class Net_CNN_11(nn.Module):
    """
        Simple CNN network.
        Four Conv1d layers
        A fc layer
        A output layer
    """
    def __init__(self, device, input_size, sequence_length, out_channel=64, padding=2, conv_kernel_size=5,
                 pool_kernel_size=2, dense_size=64, dropout=0.5):
        super(Net_CNN_11, self).__init__()
        self.device = device
        self.input_size = input_size  # the vector dimension for each AA or Nucleotide
        self.sequence_length = sequence_length  # biological sequence length
        self.conv1 = nn.Sequential(  # (input_size sequence_length)
            nn.Conv1d(in_channels=input_size, out_channels=out_channel, kernel_size=conv_kernel_size, stride=1,
                      padding=padding),
            nn.ReLU(),
            nn.MaxPool1d(kernel_size=pool_kernel_size)  # -> (out_channel, sequence_length//2)
        )
        self.dropout1 = nn.Dropout(dropout)
        self.conv2 = nn.Sequential(
            nn.Conv1d(in_channels=out_channel, out_channels=out_channel, kernel_size=conv_kernel_size, stride=1,
                      padding=padding),
            nn.ReLU(),
            nn.MaxPool1d(kernel_size=pool_kernel_size)  # -> (out_channel, sequence_length//4)
        )
        self.dropout2 = nn.Dropout(dropout)
        self.conv3 = nn.Sequential(
            nn.Conv1d(in_channels=out_channel, out_channels=out_channel, kernel_size=conv_kernel_size, stride=1,
                      padding=padding),
            nn.ReLU(),
            nn.MaxPool1d(kernel_size=pool_kernel_size)  # -> (out_channel, sequence_length//8)
        )
        self.dropout3 = nn.Dropout(dropout)
        self.conv4 = nn.Sequential(
            nn.Conv1d(in_channels=out_channel, out_channels=out_channel, kernel_size=conv_kernel_size, stride=1,
                      padding=padding),
            nn.ReLU(),
            nn.MaxPool1d(kernel_size=pool_kernel_size)  # -> (out_channel, sequence_length//16)
        )
        self.dropout4 = nn.Dropout(dropout)
        self.fc1 = nn.Linear(out_channel * self.calculate_dimension(sequence_length, padding, conv_kernel_size),
                             dense_size)
        self.fc2 = nn.Linear(dense_size, 1)

        # best model save
        self.best_model_dict = None

    def calculate_dimension(self, sequence_length, padding, kernel_size):
        L_conv1 = (sequence_length + 2 * padding - kernel_size + 1) // 2
        L_conv2 = (L_conv1 + 2 * padding - kernel_size + 1) // 2
        L_conv3 = (L_conv2 + 2 * padding - kernel_size + 1) // 2
        L_conv4 = (L_conv3 + 2 * padding - kernel_size + 1) // 2
        return L_conv4

    def forward(self, X, is_train=False, is_feature=False):
        out = X.view(-1, self.sequence_length, self.input_size)
        out = out.permute(0, 2, 1)
        out = self.conv1(out)
        if is_train:
            out = self.dropout1(out)
        out = self.conv2(out)
        if is_train:
            out = self.dropout2(out)
        out = self.conv3(out)
        if is_train:
            out = self.dropout3(out)
        out = self.conv4(out)
        if is_train:
            out = self.dropout4(out)
        out = out.view(out.size(0), -1)
        out = F.relu(self.fc1(out))
        if is_feature:
            return out
        out = torch.sigmoid(self.fc2(out))
        return out

    def fit(self, train_loader, valid_loader, epochs=1000, early_stopping=100, lr=0.001):
        # Loss and Optimizer
        criterion = nn.BCELoss()
        optimizer = torch.optim.Adam(self.parameters(), lr=lr)
        minmum_loss = np.inf
        last_saving = 0

        for epoch in range(epochs):
            time0 = time.time()
            loss_train = np.inf
            for i, (batch_X, batch_y) in enumerate(train_loader):
                Xb = batch_X.float().to(self.device)
                yb = batch_y.float().to(self.device)

                # Forward + Backward + Optimize
                optimizer.zero_grad()
                y_ = self.forward(Xb, is_train=True)
                loss = criterion(y_, yb)
                loss.backward()
                optimizer.step()
                loss_train = loss.item()
            loss_valid = float(self.evaluate(valid_loader))
            if loss_valid < minmum_loss:
                minmum_loss = loss_valid
                # torch.save(self.state_dict(), out)
                self.best_model_dict = self.state_dict()
                last_saving = epoch
            # if last_saving == epoch:
            #     print('Epoch [%d/%d]: val_loss improved to %f, saving model to model file -- this is net 11'
            #           % (epoch + 1, epochs, minmum_loss))
            # else:
            #     print('Epoch [%d/%d]: val_loss did not improve' % (epoch + 1, epochs))
            # interval = time.time() - time0
            # print('%.1fs - loss: %f - val_loss: %f\n' % (interval, loss_train, loss_valid))
            if epoch - last_saving > early_stopping: break

        if not self.best_model_dict is None:
            self.load_state_dict(self.best_model_dict)

    def evaluate(self, loader):
        loss = 0
        criterion = nn.BCELoss()
        for batch_X, batch_y in loader:
            Xb = batch_X.float().to(self.device)
            yb = batch_y.float().to(self.device)
            y_ = self.forward(Xb)
            loss += criterion(y_, yb).item()
        return loss / len(loader)

    def predict(self, loader):
        score = []
        for batch_X, batch_y in loader:
            Xb = batch_X.float().to(self.device)
            y_ = self.forward(Xb)
            score.append(y_.cpu().data)
        score = torch.cat(score).numpy()
        tmp_score = np.zeros((len(score), 2))
        tmp_score[:, 0] = 1 - score[:, 0]
        tmp_score[:, 1] = score[:, 0]
        return tmp_score


class Net_RNN_2(nn.Module):
    def __init__(self, device, category, input_size, sequence_length, hidden_size=32, num_layers=1, dense_size=64, dropout=0.5, bidirectional=False):
        super(Net_RNN_2, self).__init__()
        self.input_size = input_size
        self.sequence_length = sequence_length
        self.device = device
        self.lstm = nn.LSTM(
            input_size=input_size,
            hidden_size=hidden_size,
            num_layers=num_layers,
            batch_first=True,
            bidirectional=bidirectional,
        )
        self.dropout = nn.Dropout(dropout)
        if bidirectional:
            fc_input_size = 2 * hidden_size
        else:
            fc_input_size = hidden_size
        self.fc1 = nn.Linear(fc_input_size, dense_size)
        self.fc2 = nn.Linear(dense_size, category)

        # best model save
        self.best_model_dict = None

    def forward(self, X, is_train=False):
        out, _ = self.lstm(X, None)
        out = self.dropout(out)
        out = out[:, -1, :]
        out = F.relu(self.fc1(out))
        if is_train:
            out = self.dropout(out)
        # out = torch.softmax(self.fc2(out), dim=1)
        out = self.fc2(out)
        return out

    def fit(self, train_loader, valid_loader, epochs=1000, early_stopping=100, lr=0.001):
        # Loss and Optimizer
        criterion = nn.CrossEntropyLoss()
        optimizer = torch.optim.Adam(self.parameters(), lr=lr)
        minmum_loss = np.inf
        last_saving = 0
        for epoch in range(epochs):
            time0 = time.time()
            loss_train = np.inf
            for i, (batch_X, batch_y) in enumerate(train_loader):
                Xb = batch_X.view(-1, self.sequence_length, self.input_size).float().to(self.device)
                yb = batch_y.view(-1).long().to(self.device)
                optimizer.zero_grad()
                y_ = self.forward(Xb, is_train=True)
                loss = criterion(y_, yb)
                loss.backward()
                optimizer.step()
                loss_train = loss.item()
            loss_valid = float(self.evaluate(valid_loader))
            if loss_valid < minmum_loss:
                minmum_loss = loss_valid
                # torch.save(self.state_dict(), out)
                self.best_model_dict = self.state_dict()
                last_saving = epoch
            # if last_saving == epoch:
            #     print('Epoch [%d/%d]: val_loss improved to %f, saving model to %s'
            #           % (epoch + 1, epochs, minmum_loss, 'out'))
            # else:
            #     print('Epoch [%d/%d]: val_loss did not improve' % (epoch + 1, epochs))
            # interval = time.time() - time0
            # print('%.1fs - loss: %f - val_loss: %f\n' % (interval, loss_train, loss_valid))
            if epoch - last_saving > early_stopping: break
        if not self.best_model_dict is None:
            self.load_state_dict(self.best_model_dict)

    def evaluate(self, loader):
        loss = 0
        criterion = nn.CrossEntropyLoss()
        for batch_X, batch_y in loader:
            Xb = batch_X.view(-1, self.sequence_length, self.input_size).float().to(self.device)
            yb = batch_y.view(-1).long().to(self.device)
            y_ = self.forward(Xb)
            loss += criterion(y_, yb).item()
        return loss / len(loader)

    def predict(self, loader):
        score = []
        for batch_X, batch_y in loader:
            Xb = batch_X.view(-1, self.sequence_length, self.input_size).float().to(self.device)
            yb = batch_y.view(-1).long().to(self.device)
            y_ = self.forward(Xb)
            score.append(y_.cpu().data)
        return torch.cat(score).numpy()


class Net_ABCNN_4(nn.Module):
    def __init__(self, device, category, input_size, sequence_length, dropout=0.75):
        super(Net_ABCNN_4, self).__init__()
        self.device = device
        self.input_size = input_size                                    # the vector dimension for each AA or Nucleotide
        self.sequence_length = sequence_length                          # biological sequence length
        self.conv1 = nn.Sequential(                                     # (200 * sequence_length)
            nn.Conv1d(in_channels=input_size, out_channels=200, kernel_size=1, stride=1, padding=0),
            nn.ReLU()
        )
        self.dropout1 = nn.Dropout(dropout)
        self.conv2 = nn.Sequential(                                     # (150 * (sequence_length - 8))
            nn.Conv1d(in_channels=200, out_channels=150, kernel_size=9, stride=1, padding=0),
            nn.ReLU()
        )
        self.dropout2 = nn.Dropout(dropout)
        self.conv3 = nn.Sequential(                                     # (200 * (sequence_length - 17))
            nn.Conv1d(in_channels=150, out_channels=200, kernel_size=10, stride=1, padding=0),
            nn.ReLU()
        )
        self.dropout3 = nn.Dropout(dropout)
        #
        self.attention_1 = Attention((self.calculate_dimension(self.sequence_length, 0, 1, 9, 10), 200), 8)
        self.attention_2 = Attention((200, self.calculate_dimension(self.sequence_length, 0, 1, 9, 10)), 10)
        self.fc1 = nn.Linear(2*(self.calculate_dimension(self.sequence_length, 0, 1, 9, 10) + 200), 150)
        self.fc2 = nn.Linear(150, 8)
        self.fc3 = nn.Linear(8, category)

        #
        self.best_model_dict = None

    def calculate_dimension(self, sequence_length, padding, kernel_size_1, kernel_size_2, kernel_size_3):
        L_conv1 = sequence_length + 2 * padding - kernel_size_1 + 1
        L_conv2 = L_conv1 + 2 * padding - kernel_size_2 + 1
        L_conv3 = L_conv2 + 2 * padding - kernel_size_3 + 1
        return L_conv3

    def forward(self, X, is_train=False):
        out = X.view(-1, self.sequence_length, self.input_size)
        out = out.permute(0, 2, 1)
        out = self.conv1(out)
        if is_train:
            out = self.dropout1(out)
        out = self.conv2(out)
        if is_train:
            out = self.dropout2(out)
        out = self.conv3(out)
        if is_train:
            out = self.dropout3(out)
        out_reshape = out.permute(0, 2, 1)
        out_reshape = self.attention_1(out_reshape)
        out = self.attention_2(out)
        out_merge = torch.cat((out, out_reshape), 1)
        out_merge = F.relu(self.fc1(out_merge))
        out_merge = F.relu(self.fc2(out_merge))
        # out_merge = torch.softmax(self.fc3(out_merge), dim=1)
        out_merge = self.fc3(out_merge)
        return out_merge

    def fit(self, train_loader, valid_loader, epochs=1000, early_stopping=100, lr=0.001):
        # Loss and Optimizer
        criterion = nn.CrossEntropyLoss()
        optimizer = torch.optim.Adam(self.parameters(), lr=lr)
        minmum_loss = np.inf
        last_saving = 0
        for epoch in range(epochs):
            time0 = time.time()
            loss_train = np.inf
            for i, (batch_X, batch_y) in enumerate(train_loader):
                Xb = batch_X.float().to(self.device)
                yb = batch_y.view(-1).long().to(self.device)
                # Forward + Backward + Optimize
                optimizer.zero_grad()
                y_ = self.forward(Xb, is_train=True)
                loss = criterion(y_, yb)
                loss.backward()
                optimizer.step()
                loss_train = loss.item()
            loss_valid = float(self.evaluate(valid_loader))
            if loss_valid < minmum_loss:
                minmum_loss = loss_valid
                # torch.save(self.state_dict(), out)
                self.best_model_dict = self.state_dict()
                last_saving = epoch
            # if last_saving == epoch:
            #     print('Epoch [%d/%d]: val_loss improved to %f, saving model to %s'
            #           % (epoch + 1, epochs, minmum_loss, 'out'))
            # else:
            #     print('Epoch [%d/%d]: val_loss did not improve' % (epoch + 1, epochs))
            # interval = time.time() - time0
            # print('%.1fs - loss: %f - val_loss: %f\n' % (interval, loss_train, loss_valid))
            if epoch - last_saving > early_stopping: break
        if not self.best_model_dict is None:
            self.load_state_dict(self.best_model_dict)

    def evaluate(self, loader):
        loss = 0
        criterion = nn.CrossEntropyLoss()
        for batch_X, batch_y in loader:
            Xb = batch_X.float().to(self.device)
            yb = batch_y.view(-1).long().to(self.device)
            y_ = self.forward(Xb)
            loss += criterion(y_, yb).item()
        return loss / len(loader)

    def predict(self, loader):
        score = []
        for batch_X, batch_y in loader:
            Xb = batch_X.float().to(self.device)
            yb = batch_y.view(-1).long().to(self.device)
            y_ = self.forward(Xb)
            score.append(y_.cpu().data)
        return torch.cat(score).numpy()


class Net_ResNet_5(nn.Module):
    def __init__(self, device, category, input_size, sequence_length, out_channel=64, dense_size=64):
        super(Net_ResNet_5, self).__init__()
        self.device = device
        self.input_size = input_size
        self.sequence_length = sequence_length
        self.out_channel = out_channel
        self.dense_size = dense_size
        # The first layer
        self.pre = nn.Sequential(
            nn.Conv1d(input_size, out_channel, kernel_size=7, stride=2, padding=3, bias=False),
            nn.BatchNorm1d(out_channel),
            nn.ReLU(inplace=True),
            nn.MaxPool1d(2)
        )
        # Repeated layers
        self.layer1 = self._make_layer(out_channel, out_channel, 2, 1)
        self.layer2 = self._make_layer(out_channel, 2*out_channel, 2, 2)
        self.layer3 = self._make_layer(2*out_channel, 4*out_channel, 2, 2)
        self.layer4 = self._make_layer(4*out_channel, 8*out_channel, 2, 2)

        # FC layer
        self.fc1 = nn.Linear(self.calculate_dimension(self.sequence_length) * 8 * out_channel, dense_size)
        self.fc2 = nn.Linear(dense_size, category)

        #
        self.best_model_dict = None

    def _make_layer(self, in_channel, out_channel, block_num, stride=1):
        shortcut = nn.Sequential(
            nn.Conv1d(in_channel, out_channel, kernel_size=1, stride=stride, bias=False),
            nn.BatchNorm1d(out_channel)
        )
        layers = []
        layers.append(ResidualBlock(in_channel, out_channel, stride, shortcut))

        for i in range(1, block_num):
            layers.append(ResidualBlock(out_channel, out_channel))

        return nn.Sequential(*layers)

    def forward(self, X):
        out = X.view(-1, self.sequence_length, self.input_size)
        out = out.permute(0, 2, 1)
        # print('out: ', out.size())
        out = self.pre(out)
        # print('Pre: ', out.size())
        out = self.layer1(out)
        # print('layer 1: ', out.size())
        out = self.layer2(out)
        # print('layer 2: ', out.size())
        out = self.layer3(out)
        # print('layer 3: ', out.size())
        out = self.layer4(out)
        # print('layer 4: ', out.size())
        out = F.avg_pool1d(out, 2)
        out = out.view(out.size(0), -1)
        out = F.relu(self.fc1(out))
        # out = torch.softmax(self.fc2(out), dim=1)
        out = self.fc2(out)
        return out

    def fit(self, train_loader, valid_loader, epochs=1000, early_stopping=100, lr=0.001):
        # Loss and Optimizer
        criterion = nn.CrossEntropyLoss()
        optimizer = torch.optim.Adam(self.parameters(), lr=lr)
        minmum_loss = np.inf
        last_saving = 0

        for epoch in range(epochs):
            time0 = time.time()
            loss_train = np.inf
            for i, (batch_X, batch_y) in enumerate(train_loader):
                Xb = batch_X.float().to(self.device)
                yb = batch_y.view(-1).long().to(self.device)

                # Forward + Backward + Optimize
                optimizer.zero_grad()
                y_ = self.forward(Xb)
                loss = criterion(y_, yb)
                loss.backward()
                optimizer.step()
                loss_train = loss.item()
            loss_valid = float(self.evaluate(valid_loader))
            if loss_valid < minmum_loss:
                minmum_loss = loss_valid
                # torch.save(self.state_dict(), out)
                self.best_model_dict = self.state_dict()
                last_saving = epoch
            # if last_saving == epoch:
            #     print('Epoch [%d/%d]: val_loss improved to %f, saving model to %s'
            #           % (epoch + 1, epochs, minmum_loss, 'out'))
            # else:
            #     print('Epoch [%d/%d]: val_loss did not improve' % (epoch + 1, epochs))
            # interval = time.time() - time0
            # print('%.1fs - loss: %f - val_loss: %f\n' % (interval, loss_train, loss_valid))
            if epoch - last_saving > early_stopping: break
        if not self.best_model_dict is None:
            self.load_state_dict(self.best_model_dict)

    def evaluate(self, loader):
        loss = 0
        criterion = nn.CrossEntropyLoss()
        for batch_X, batch_y in loader:
            Xb = batch_X.float().to(self.device)
            yb = batch_y.view(-1).long().to(self.device)
            y_ = self.forward(Xb)
            loss += criterion(y_, yb).item()
        return loss / len(loader)

    def predict(self, loader):
        score = []
        for batch_X, batch_y in loader:
            Xb = batch_X.float().to(self.device)
            yb = batch_y.view(-1).to(self.device)
            y_ = self.forward(Xb)
            score.append(y_.cpu().data)
        return torch.cat(score).numpy()

    def calculate_dimension(self, sequence_length):
        pre_out = ((sequence_length - 1) // 2 + 1) // 2
        layer1_out = pre_out
        layer2_out = (layer1_out - 1) // 2 + 1
        layer3_out = (layer2_out - 1) // 2 + 1
        layer4_out = (layer3_out - 1) // 2 + 1
        return layer4_out // 2


class Net_AutoEncoder_6(nn.Module):
    def __init__(self, device, category, input_dim):
        super(Net_AutoEncoder_6, self).__init__()
        self.device = device
        self.input_dim = input_dim
        self.category = category
        self.best_model_dict = None
        self.best_classification_model_dict = None

        self.encoder = nn.Sequential(
            nn.Linear(input_dim, 256),
            nn.ReLU(),
            nn.Linear(256, 128)
        )

        self.decoder = nn.Sequential(
            nn.Linear(128, 256),
            nn.ReLU(),
            nn.Linear(256, input_dim)
        )

    def forward(self, X, re_train=False):
        encoder = self.encoder(X)
        if re_train:
            decoder = torch.softmax(self.decoder(encoder), dim=1)
        else:
            decoder = self.decoder(encoder)
        return decoder

    def fit(self, train_loader, valid_loader, epochs=1000, early_stopping=100, lr=0.001):
        criterion = nn.MSELoss()
        optimizer = torch.optim.Adam(self.parameters(), lr=lr)
        minmum_loss = np.inf
        last_saving = 0
        for epoch in range(epochs):
            time0 = time.time()
            loss_train = np.inf
            for i, (batch_X, batch_y) in enumerate(train_loader):
                Xb = batch_X.float().to(self.device)
                yb = batch_X.float().to(self.device)
                # Forward + Backward + Optimize
                optimizer.zero_grad()
                y_ = self.forward(Xb)
                loss = criterion(y_, yb)
                loss.backward()
                optimizer.step()
                loss_train = loss.item()
            loss_valid = float(self.evaluate(valid_loader))
            if loss_valid < minmum_loss:
                minmum_loss = loss_valid
                self.best_model_dict = self.state_dict()
                last_saving = epoch
            if epoch - last_saving > early_stopping: break
        self.load_state_dict(self.best_model_dict)

    def evaluate(self, loader):
        loss = 0
        criterion = nn.MSELoss()
        for batch_X, batch_y in loader:
            Xb = batch_X.float().to(self.device)
            yb = batch_X.float().to(self.device)
            y_ = self.forward(Xb)
            loss += criterion(y_, yb).item()
        return loss / len(loader)

    def re_build_net(self):
        self.load_state_dict(self.best_model_dict)
        self.decoder = nn.Linear(128, self.category).to(self.device)

        for name, param in self.named_parameters():
            if 'decoder' in name:
                param.requires_grad = True
            else:
                param.requires_grad = False

    def re_fit(self, train_loader, valid_loader, epochs=1000, early_stopping=100, lr=0.001):
        # Loss and Optimizer
        criterion = nn.CrossEntropyLoss()
        optimizer = torch.optim.Adam(self.parameters(), lr=lr)
        minmum_loss = np.inf
        last_saving = 0
        for epoch in range(epochs):
            time0 = time.time()
            loss_train = np.inf
            for i, (batch_X, batch_y) in enumerate(train_loader):
                Xb = batch_X.float().to(self.device)
                yb = batch_y.view(-1).long().to(self.device)
                # Forward + Backward + Optimize
                optimizer.zero_grad()
                y_ = self.forward(Xb, re_train=True)
                loss = criterion(y_, yb)
                loss.backward()
                optimizer.step()
                loss_train = loss.item()
            loss_valid = float(self.re_evaluate(valid_loader))
            if loss_valid < minmum_loss:
                minmum_loss = loss_valid
                # torch.save(self.state_dict(), out)
                self.best_classification_model_dict = self.state_dict()
                last_saving = epoch
            if epoch - last_saving > early_stopping: break
        if not self.best_model_dict is None:
            self.load_state_dict(self.best_classification_model_dict)

    def re_evaluate(self, loader):
        loss = 0
        criterion = nn.CrossEntropyLoss()
        for batch_X, batch_y in loader:
            Xb = batch_X.float().to(self.device)
            yb = batch_y.view(-1).long().to(self.device)
            y_ = self.forward(Xb, re_train=True)
            loss += criterion(y_, yb).item()
        return loss / len(loader)

    def predict(self, loader):
        score = []
        for batch_X, batch_y in loader:
            Xb = batch_X.float().to(self.device)
            yb = batch_y.view(-1).long().to(self.device)
            y_ = self.forward(Xb, re_train=True)
            score.append(y_.cpu().data)
        return torch.cat(score).numpy()




#!/usr/bin/env python
# _*_ coding: utf-8 _*_

import numpy as np
import pandas as pd
import copy

class ModelMetrics(object):
    def __init__(self):
        super(ModelMetrics, self).__init__()
        self.metrics = pd.DataFrame(data=None, columns=['Id', 'StartTime', 'EndTime', 'Sn', 'Sp', 'Pre', 'Acc', 'MCC', 'F1', 'AUROC', 'AUPRC'])
        self.classification_task = None
        self.aucData = {}                                               # dict key=Id, value=Pandas DataFrame
        self.prcData = {}                                               # dict key=Id, value=Pandas DataFrame
        self.prediction_scores = {}                                     # dict key=Id, value=Pandas DataFrame
        self.models = {}                                                # dict key=Id, value=model_list

    def insert_data(self, metric_ndarray, id, auc_df, prc_df, prediction_score, model=None):
        df = pd.DataFrame(data=metric_ndarray)
        self.metrics = self.metrics.append(df, ignore_index=True)
        self.aucData[id] = auc_df
        self.prcData[id] = prc_df
        self.prediction_scores[id] = prediction_score
        if not model is None:
            self.models[id] = copy.deepcopy(model)


if __name__ == '__main__':
    model = ModelMetrics()
    df = pd.DataFrame([['id', '000', '111', 98, 99, 98, 99, 0.9, 0.8, 0.90, 0.91]], columns=['Id', 'StartTime', 'EndTime', 'Sn', 'Sp', 'Pre', 'Acc', 'MCC', 'F1', 'AUROC', 'AUPRC'])
    df1 = pd.DataFrame([['id', '111', '111', 98, 99, 98, 99, 0.9, 0.8, 0.90, 0.91]], columns=['Id', 'StartTime', 'EndTime', 'Sn', 'Sp', 'Pre', 'Acc', 'MCC', 'F1', 'AUROC', 'AUPRC'])

    model.insert_data(df, None, None)
    model.insert_data(df1, None, None)


